#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""SEGWO package."""

import importlib.metadata

from . import cov, response, sensitivity, sym

__copyright__ = "2025, Olaf Hartwig and Jean-Baptiste Bayle"
__author__ = "Olaf Hartwig and Jean-Baptiste Bayle"
__version__ = importlib.metadata.version("segwo")
